#include "precomp.h"

extern struct file_operations aslcxp_ops;

static struct aslcxp_driver_struct {
   dev_t         driver_dev;
   struct class *class;
   dev_t         next_minor;
   spinlock_t    lock;
} aslcxp_driver;

static dev_t aslcxp_driver_get_next_minor(struct aslcxp_device *cxp_device)
{
   dev_t minor = 0;

   spin_lock(&aslcxp_driver.lock);
   minor = aslcxp_driver.next_minor++;
   spin_unlock(&aslcxp_driver.lock);

   return minor;
}

static int aslcxp_device_probe(struct mda_device *mdadev, const struct mda_device_id *template)
{
   int error = -ENXIO;
   struct aslcxp_device *cxp_device = NULL;

   cxp_device = kzalloc(sizeof(*cxp_device), GFP_KERNEL);
   if (cxp_device == NULL){
      printk(KERN_ERR "%s: kzalloc (device) failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_kzalloc_device;
   }

   cxp_device->mdadev = mdadev;
   cxp_device->major = MAJOR(aslcxp_driver.driver_dev);
   cxp_device->minor =  aslcxp_driver_get_next_minor(cxp_device);
   cxp_device->owner = THIS_MODULE;
   cxp_device->cxp_protocol_version = cxp_unknown;
   
   atomic_set(&cxp_device->log_count, 0);

   mutex_init(&cxp_device->open_mutex);
   spin_lock_init(&cxp_device->interrupt_spinlock);
   mutex_init(&cxp_device->cxp_serialisation_mutex);

   cxp_device->cxp_transaction.cxp_device = cxp_device;
   cxp_device->cxp_transaction.cxp_transaction_state = cxp_idle;
   init_waitqueue_head(&cxp_device->cxp_transaction.wait_queue_head);
   aslcxp_init_ack_timer(&cxp_device->cxp_transaction.ack_timer, aslcxp_timeout_callback);

   mda_set_drvdata(mdadev, cxp_device);

   error = mda_interrupt_register(mdadev,
                                  mdadev->params.interrupt_register,
                                  mdadev->params.interrupt_mask,
                                  aslcxp_isr,
                                  cxp_device);
   if (error){
      printk(KERN_ERR "%s: mda_interrupt_register returned %d\n", __FUNCTION__, error);
      goto err_irq;
   }

   cdev_init(&cxp_device->cdev, &aslcxp_ops);
   cxp_device->cdev.owner = THIS_MODULE;
   error = cdev_add(&cxp_device->cdev, MKDEV(cxp_device->major, cxp_device->minor), 1);
   if (error){
      printk(KERN_ERR "%s: cdev_add failed with %d\n", __FUNCTION__, error);
      goto err_cdev;
   }

   cxp_device->device = device_create(aslcxp_driver.class,
                                    &mdadev->device,
                                    MKDEV(cxp_device->major, cxp_device->minor),
                                    NULL,
                                    "%s%d", ASLCXP_DEVICE_NAME, cxp_device->minor);

   if (cxp_device->device == NULL){
      printk(KERN_ERR "%s: device_create failed\n", __FUNCTION__);
      error = -ENODEV;
      goto err_device;
   }

   sprintf(mdadev->name, "%s", dev_name(cxp_device->device));

   dev_printk(KERN_INFO, cxp_device->device, "registered at %u/%u (%s) with timeout granularity %ums\n",
                         cxp_device->major, cxp_device->minor, mdadev->name, jiffies_to_msecs(1));

   error = 0;
   goto finished;

err_device:
   cdev_del(&cxp_device->cdev);
err_cdev:
   error = mda_interrupt_unregister(mdadev,
                                    mdadev->params.interrupt_register,
                                    mdadev->params.interrupt_mask);
err_irq:
   kfree(cxp_device);
err_kzalloc_device:
finished:
   return error;
}

static void aslcxp_device_remove(struct mda_device *mdadev)
{
   struct aslcxp_device *cxp_device = mda_get_drvdata(mdadev);

   dev_printk(KERN_INFO, cxp_device->device, "%s: unloading\n", __FUNCTION__);
   device_destroy(aslcxp_driver.class, MKDEV(cxp_device->major, cxp_device->minor));
   cdev_del(&cxp_device->cdev);
   mda_interrupt_unregister(mdadev,
                            mdadev->params.interrupt_register,
                            mdadev->params.interrupt_mask);
   kfree(cxp_device);
   return;
}

static int aslcxp_device_suspend(struct mda_device *mdadev, pm_message_t state)
{
   struct aslcxp_device *cxp_device = mda_get_drvdata(mdadev);

   if (cxp_device){
      dev_printk(KERN_INFO, cxp_device->device, "%s\n", __FUNCTION__);
   }
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static int aslcxp_device_suspend_late(struct mda_device *mdadev, pm_message_t state)
{
   struct aslcxp_device *cxp_device = mda_get_drvdata(mdadev);

   if (cxp_device){
      dev_printk(KERN_INFO, cxp_device->device, "%s\n", __FUNCTION__);
   }
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static int aslcxp_device_resume(struct mda_device *mdadev)
{
   struct aslcxp_device *cxp_device = mda_get_drvdata(mdadev);

   if (cxp_device) dev_printk(KERN_INFO, cxp_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static int aslcxp_device_resume_early(struct mda_device *mdadev)
{
   struct aslcxp_device *cxp_device = mda_get_drvdata(mdadev);

   if (cxp_device) dev_printk(KERN_INFO, cxp_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static void aslcxp_device_shutdown(struct mda_device *mdadev)
{
   struct aslcxp_device *cxp_device = mda_get_drvdata(mdadev);

   if (!cxp_device){
      printk(KERN_INFO "%s - no device\n", __FUNCTION__);
      goto bail;
   }

   dev_printk(KERN_INFO, cxp_device->device, "%s\n", __FUNCTION__);

bail:
   return;
}

struct mda_device_id aslcxp_mda_table[] =
{
   {
      "aslcxp",
      0
   },
   { 0 }                                       /* Terminate with empty struct */
};

struct mda_driver aslcxp_mda_driver = {
   .name         = ASLCXP_DEVICE_NAME,
   .probe        = aslcxp_device_probe,
   .remove       = aslcxp_device_remove,
   .suspend      = aslcxp_device_suspend,
   .suspend_late = aslcxp_device_suspend_late,
   .resume       = aslcxp_device_resume,
   .resume_early = aslcxp_device_resume_early,
   .shutdown     = aslcxp_device_shutdown,
   .id_table     = aslcxp_mda_table,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
static CLASS_ATTR_STRING(version, S_IRUGO, ASLCXP_VERSION_STRING);
#endif

static int __init aslcxp_driver_init(void)
{
   int error = 0;

   printk(KERN_INFO "%s: Built on %s at %s\n", __FUNCTION__, __DATE__, __TIME__);

   memset(&aslcxp_driver, 0, sizeof(aslcxp_driver));

   spin_lock_init(&aslcxp_driver.lock);

   aslcxp_driver.class = class_create(THIS_MODULE, ASLCXP_DEVICE_NAME);
   if (!aslcxp_driver.class){
      error = -ENOMEM;
      printk(KERN_ERR "%s class_create returned %d\n", __FUNCTION__, error);
      goto err_class;
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   error = class_create_file(aslcxp_driver.class, &class_attr_version.attr);
   if (error)
   {
      printk(KERN_ERR "%s class_create_file returned %d\n", __FUNCTION__, error);
      goto err_attr;
   }
#endif
   error = alloc_chrdev_region(&aslcxp_driver.driver_dev, 0, ASLCXP_MAX_DEVICES, ASLCXP_DEVICE_NAME);
   if (error){
      printk(KERN_ERR "%s alloc_chrdev_region returned %d\n", __FUNCTION__, error);
      goto err_chr;
   }

   printk(KERN_INFO "%s: major number = %u\n", __FUNCTION__, MAJOR(aslcxp_driver.driver_dev));
   error = mda_register_driver(&aslcxp_mda_driver);
   if (error){
      printk(KERN_ERR "%s mda_register_driver returned %d\n", __FUNCTION__, error);
      goto err_mda;
   }

   goto finished;

err_mda:
   unregister_chrdev_region(aslcxp_driver.driver_dev, ASLCXP_MAX_DEVICES);
err_chr:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   class_remove_file(aslcxp_driver.class,  &class_attr_version.attr);
err_attr:
#endif
   class_destroy(aslcxp_driver.class);
err_class:
finished:
   return error;
}

static void __exit aslcxp_driver_cleanup(void)
{
   printk(KERN_INFO "%s\n", __FUNCTION__);

   mda_unregister_driver(&aslcxp_mda_driver);
   unregister_chrdev_region(aslcxp_driver.driver_dev, ASLCXP_MAX_DEVICES);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   class_remove_file(aslcxp_driver.class,  &class_attr_version.attr);
#endif
   class_destroy(aslcxp_driver.class);

   return;
}

module_init(aslcxp_driver_init);
module_exit(aslcxp_driver_cleanup);
MODULE_DEVICE_TABLE(mda, aslcxp_mda_table);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Chris Read <techsupport@activesilicon.com>");
