#include "precomp.h"

extern struct file_operations asldma_ops;

#if defined __x86_64 || defined __x86_64__ || defined __amd64 || defined __amd64__
int asldma_enable_dma_sync = 0;
#else
int asldma_enable_dma_sync = 1;
#endif

static struct asldma_driver_struct {
   struct class * class;
   dev_t          driver_dev;
   dev_t          next_minor;
   spinlock_t     lock;
} asldma_driver;

static dev_t asldma_driver_get_next_minor(struct asldma_device *dma_device)
{
   dev_t minor = 0;

   spin_lock(&asldma_driver.lock);
   minor = asldma_driver.next_minor++;
   spin_unlock(&asldma_driver.lock);

   return minor;
}

static int asldma_device_probe(struct mda_device *mdadev, const struct mda_device_id *template)
{
   int error = -ENXIO;
   struct asldma_device *dma_device = NULL;
   int event_index = 0;
   struct device * base_device = mdadev->parent->parent->parent->device;

   dma_device = kzalloc(sizeof(*dma_device), GFP_KERNEL);
   if (dma_device == NULL){
      dev_err(base_device, "%s: kzalloc (device) failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_kzalloc_device;
   }
   dma_device->event_list_size = ASLDMA_MAX_EVENTS * sizeof(struct asldma_event);
   dma_device->events = kzalloc(dma_device->event_list_size, GFP_KERNEL);
   if (dma_device->events == NULL){
      dev_err(base_device, "%s: kzalloc (events) failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_kzalloc_event;
   }

   dma_device->mdadev = mdadev;

   dma_device->major = MAJOR( asldma_driver.driver_dev );
   dma_device->minor = asldma_driver_get_next_minor(dma_device);
   dma_device->owner = THIS_MODULE;

   dma_device->baseboard_type = mda_get_baseboard_type(mdadev);

   spin_lock_init(&dma_device->open_lock);
   spin_lock_init(&dma_device->event_list_lock);
   spin_lock_init(&dma_device->dma_buffer_list_lock);

   init_waitqueue_head(&dma_device->event_wait_queue_head);

   INIT_LIST_HEAD(&dma_device->dma_buffer_list);
   INIT_LIST_HEAD(&dma_device->event_free_list);
   INIT_LIST_HEAD(&dma_device->event_pending_list);

   for (event_index = 0;
        event_index < dma_device->event_list_size/sizeof(struct asldma_event);
        event_index++){
      INIT_LIST_HEAD(&dma_device->events[event_index].this_event);
      list_add_tail(&dma_device->events[event_index].this_event, &dma_device->event_free_list);
   }

   mda_set_drvdata(mdadev, dma_device);

   error = mda_interrupt_register(mdadev,
                                  mdadev->params.interrupt_register,
                                  mdadev->params.interrupt_mask,
                                  asldma_isr,
                                  dma_device);
   if (error){
      dev_err(base_device, "%s: mda_interrupt_register returned %d\n", __FUNCTION__, error);
      goto err_irq;
   }

   cdev_init(&dma_device->cdev, &asldma_ops);
   dma_device->cdev.owner = THIS_MODULE;

   error = cdev_add(&dma_device->cdev, MKDEV(dma_device->major, dma_device->minor), 1);
   if (error){
      dev_err(base_device, "%s: cdev_add failed with %d\n", __FUNCTION__, error);
      goto err_cdev;
   }

   dma_device->device = device_create(asldma_driver.class,
                                      &mdadev->device,
                                      MKDEV(dma_device->major, dma_device->minor),
                                      NULL,
                                      "%s%d",
                                      ASLDMA_DEVICE_NAME,
                                      dma_device->minor);

   if (dma_device->device == NULL){
      dev_err(base_device, "%s: device_create failed\n", __FUNCTION__);
      error = -ENODEV;
      goto err_device;
   }

   sprintf(mdadev->name, "%s", dev_name(dma_device->device));

   dev_printk(KERN_INFO,
              dma_device->device,
              "registered at %u/%u (%s)\n",
              dma_device->major,
              dma_device->minor,
              mdadev->name);

   error = 0;
   goto finished;

err_device:
   cdev_del(&dma_device->cdev);
err_cdev:
   error = mda_interrupt_unregister(mdadev,
                                    mdadev->params.interrupt_register,
                                    mdadev->params.interrupt_mask);
err_irq:
   kfree(dma_device->events);
err_kzalloc_event:
   kfree(dma_device);
err_kzalloc_device:
finished:
   return error;
}

void asldma_device_remove(struct mda_device *mdadev)
{
   struct asldma_device *dma_device = mda_get_drvdata(mdadev);

   dev_printk(KERN_INFO, dma_device->device, "%s: unloading\n", __FUNCTION__);
   asldma_stop_risc_engine(dma_device);
   device_destroy(asldma_driver.class, MKDEV(dma_device->major, dma_device->minor));
   cdev_del(&dma_device->cdev);
   mda_interrupt_unregister(mdadev,
                            mdadev->params.interrupt_register,
                            mdadev->params.interrupt_mask);
   kfree(dma_device->events);
   kfree(dma_device);
   return;
}

static int asldma_do_device_suspend(struct mda_device *mdadev, const char * szWhence)
{
   int error = -ENXIO;
   struct asldma_device *dma_device = NULL;

   if (!mdadev) goto bail;
   dma_device = mda_get_drvdata(mdadev);
   if (!dma_device) goto bail;

   dev_printk(KERN_INFO, dma_device->device, "%s from %s\n", __FUNCTION__, szWhence);
   asldma_stop_risc_engine(dma_device);
   error = 0;
bail:
   if (error) printk(KERN_ERR "%s NULL parameter from %s!!\n", __FUNCTION__, szWhence);
   return error;
}

static int asldma_do_device_resume(struct mda_device *mdadev, const char * szWhence)
{
   int error = -ENXIO;
   struct asldma_device *dma_device = NULL;

   if (!mdadev) goto bail;
   dma_device = mda_get_drvdata(mdadev);
   if (!dma_device) goto bail;

   dev_printk(KERN_INFO, dma_device->device, "%s from %s\n", __FUNCTION__, szWhence);
   error = 0;
bail:
   if (error) printk(KERN_ERR "%s NULL parameter from %s!!\n", __FUNCTION__, szWhence);
   return error;
}
static int asldma_device_suspend(struct mda_device *mdadev, pm_message_t state)
{
   return asldma_do_device_suspend(mdadev, __FUNCTION__);
}

static int asldma_device_suspend_late(struct mda_device *mdadev, pm_message_t state)
{
   return asldma_do_device_suspend(mdadev, __FUNCTION__);
}

static int asldma_device_resume(struct mda_device *mdadev)
{
   return asldma_do_device_resume(mdadev, __FUNCTION__);
}

static int asldma_device_resume_early(struct mda_device *mdadev)
{
   return asldma_do_device_resume(mdadev, __FUNCTION__);
}

static void asldma_device_shutdown(struct mda_device *mdadev)
{
   (void)asldma_do_device_suspend(mdadev, __FUNCTION__);

   return;
}

struct mda_device_id asldma_mda_table[] =
{
   {
      "asldma",
      0
   },
   { 0 }                                       /* Terminate with empty struct */
};

struct mda_driver asldma_mda_driver = {
   .name         = ASLDMA_DEVICE_NAME,
   .probe        = asldma_device_probe,
   .remove       = asldma_device_remove,
   .suspend      = asldma_device_suspend,
   .suspend_late = asldma_device_suspend_late,
   .resume       = asldma_device_resume,
   .resume_early = asldma_device_resume_early,
   .shutdown     = asldma_device_shutdown,
   .id_table     = asldma_mda_table,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
static CLASS_ATTR_STRING(version, S_IRUGO, ASLDMA_VERSION_STRING);
#endif

static int asldma_driver_init(void) 
{
   int error = 0;

   printk(KERN_INFO "%s: Built on %s at %s\n", __FUNCTION__, __DATE__, __TIME__);

   memset(&asldma_driver, 0, sizeof(asldma_driver));
   spin_lock_init(&asldma_driver.lock);

   asldma_driver.class = class_create(THIS_MODULE, ASLDMA_DEVICE_NAME);
   if (!asldma_driver.class){
      error = -ENOMEM;
      printk(KERN_ERR "%s class_create returned %d\n", __FUNCTION__, error);
      goto err_class;
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   error = class_create_file(asldma_driver.class, &class_attr_version.attr);
   if (error)
   {
      printk(KERN_ERR "%s class_create_file returned %d\n", __FUNCTION__, error);
      goto err_attr;
   }
#endif
   error = alloc_chrdev_region(&asldma_driver.driver_dev, 0, ASLDMA_MAX_DEVICES, ASLDMA_DEVICE_NAME);
   if (error){
      printk(KERN_ERR "%s alloc_chrdev_region returned %d\n", __FUNCTION__, error);
      goto err_chr;
   }

   printk(KERN_INFO "%s: major number = %u\n", __FUNCTION__, MAJOR(asldma_driver.driver_dev));
   error = mda_register_driver(&asldma_mda_driver);
   if (error){
      printk(KERN_ERR "%s mda_register_driver returned %d\n", __FUNCTION__, error);
      goto err_mda;
   }

   goto finished;

err_mda:
   unregister_chrdev_region(asldma_driver.driver_dev, ASLDMA_MAX_DEVICES);
err_chr:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   class_remove_file(asldma_driver.class, &class_attr_version.attr);
#endif
err_attr:
   class_destroy(asldma_driver.class);
err_class:
finished:
   return error;
}

void asldma_driver_cleanup(void)
{
   printk(KERN_INFO "%s\n", __FUNCTION__);

   mda_unregister_driver(&asldma_mda_driver);
   unregister_chrdev_region(asldma_driver.driver_dev, ASLDMA_MAX_DEVICES);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   class_remove_file(asldma_driver.class, &class_attr_version.attr);
#endif
   class_destroy(asldma_driver.class);
   return;
}
module_init(asldma_driver_init);
module_exit(asldma_driver_cleanup);
MODULE_DEVICE_TABLE(mda, asldma_mda_table);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Chris Read <techsupport@activesilicon.com>");

module_param(asldma_enable_dma_sync, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
MODULE_PARM_DESC(asldma_enable_dma_sync, "Set to enable programmitc flush and invalidate of DMA areas");

