#include "precomp.h"
#include "boards.h"

static int mda_bus_uevent    (struct device *, struct kobj_uevent_env *);
static int mda_child_uevent  (struct device *, struct kobj_uevent_env *);
static int mda_channel_uevent(struct device *, struct kobj_uevent_env *);
static struct mda_device * mda_verify_child_device(struct device * device);

struct mda_device *mda_dev_get(struct mda_device *mda_dev_in)
{
   struct device * device = NULL;
   struct mda_device *mda_dev_out = NULL;

   if (mda_dev_in) device = get_device(&mda_dev_in->device);
   if (device) mda_dev_out = to_mda_device(device);

   return mda_dev_out;
}

void mda_dev_put(struct mda_device *mda_dev)
{
   if (mda_dev) put_device(&mda_dev->device);
   return;
}

struct mda_root_node *mda_root_node_get(struct mda_root_node *mda_root_node_in)
{
   struct device * device = NULL;
   struct mda_root_node *mda_root_node_out = NULL;

   if (mda_root_node_in) device = get_device(&mda_root_node_in->device);
   if (device) mda_root_node_out = to_mda_root_node(device);

   return mda_root_node_out;
}

void mda_root_node_put(struct mda_root_node *mda_root_node)
{
   if (mda_root_node) put_device(&mda_root_node->device);
   return;
}

struct mda_channel *mda_channel_get(struct mda_channel *mda_channel_in)
{
   struct device * device = NULL;
   struct mda_channel *mda_channel_out = NULL;

   if (mda_channel_in) device = get_device( &mda_channel_in->device );
   if (device) mda_channel_out = to_mda_channel(device);

   return mda_channel_out;
}

void mda_channel_put(struct mda_channel *mda_channel)
{
   if (mda_channel) put_device( &mda_channel->device );
   return;
}

struct aslenum_device * mda_baseboard_get(struct aslenum_device *baseboard_in)
{
   struct device * device = NULL;
   struct pci_dev * pci_dev_out = NULL;

   if (baseboard_in) device = get_device(baseboard_in->device);
   if (device) pci_dev_out = pci_dev_get(baseboard_in->pcidev);
   if (device && !pci_dev_out) put_device(baseboard_in->device);

   return (device && pci_dev_out) ? baseboard_in : NULL;
}

void mda_baseboard_put(struct aslenum_device *baseboard)
{
   if (baseboard) put_device(baseboard->device);
   if (baseboard) pci_dev_put(baseboard->pcidev);
   return;
}

static const struct mda_device_id * mda_match_device(struct mda_driver *drv, struct mda_device *dev)
{
   const struct mda_device_id *id = drv->id_table;

   while(id->driver_matching_name && id->driver_matching_name[0]){
      if (0 == strncmp(id->driver_matching_name, dev->id.driver_matching_name, strlen(id->driver_matching_name))) break;
      id++;
   }
   return id->driver_matching_name ? id : NULL;
}

void mda_set_drvdata(struct mda_device *mda_device, void * data)
{
   dev_set_drvdata( &mda_device->device, data );
   return;
}

void * mda_get_drvdata(struct mda_device *mda_device)
{
   return dev_get_drvdata( &mda_device->device );
}

int mda_bus_match(struct device * device, struct device_driver * drv)
{
   struct mda_device * mda_device = mda_verify_child_device(device);

   if (mda_device){
      struct mda_driver *mda_driver = to_mda_driver(drv);
      const struct mda_device_id *found_id;

      found_id = mda_match_device(mda_driver, mda_device);
      if (found_id)
         return 1;
   }
   return 0;
}

static int mda_do_device_probe(struct mda_driver *mda_drv, struct mda_device *mda_dev)
{
   int error = -ENODEV;

   if (!mda_drv->probe || !mda_drv->id_table) goto bail;
   mda_dev->driver = mda_drv;

   error = mda_drv->probe(mda_dev, mda_match_device(mda_drv, mda_dev));
   if (error) {
      mda_dev->driver = NULL;
      dev_set_drvdata( &mda_dev->device, NULL );
   }
bail:
   return error;
}

int mda_bus_probe(struct device * device)
{
   int error = 0;
   struct mda_device * mda_device = mda_verify_child_device(device);

   if (mda_device){
      struct mda_driver *mda_driver = NULL;

      mda_driver = to_mda_driver(device->driver);
      mda_device = to_mda_device(device);
      mda_device = mda_dev_get(mda_device);
      if (mda_device == NULL){
         error = -ESRCH;
         goto bail;
      }
      error = mda_do_device_probe(mda_driver, mda_device);
      if (error){
         mda_dev_put(mda_device);
         goto bail;
      }
   }
bail:
   return error;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
void mda_bus_remove(struct device * device)
#else
int mda_bus_remove(struct device * device)
#endif
{
   struct mda_device * mda_device = mda_verify_child_device(device);

   if (mda_device){
      struct mda_driver * mda_driver = mda_device->driver;

      if (mda_driver) {
         if (mda_driver->remove) {
            mda_driver->remove(mda_device);
         }
         mda_device->driver = NULL;
      }
      pm_runtime_disable(device);
      pm_runtime_set_suspended(device);
      pm_runtime_put_noidle(device);
      mda_dev_put(mda_device);
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
#else
   return 0;
#endif
}

void mda_bus_shutdown(struct device * device)
{
   struct mda_device *mda_device = mda_verify_child_device(device);  /* Called for all bus device types */

   if (mda_device){ /* It is a child device */
      struct mda_driver *mda_driver = mda_device->driver;
      if (mda_driver){
         if (mda_driver->shutdown){
            mda_driver->shutdown(mda_device);
         }
      }
   }

   return;
}

struct bus_type mda_bus_type = {
   .name      = "mda",
   .match     = mda_bus_match,
   .uevent    = mda_bus_uevent,
   .probe     = mda_bus_probe,
   .remove    = mda_bus_remove,
   .shutdown  = mda_bus_shutdown,
#if defined _CLR_PM_TODO
   .dev_attrs = mda_bus_device_attrs,
   .bus_attrs = mda_bus_bus_attrs,
   .pm        = mda_bus_pm_ops
#endif
};

__init int aslenum_bus_init(void)
{
   return bus_register(&mda_bus_type);
}

void aslenum_bus_remove(void)
{
   bus_unregister(&mda_bus_type);
   return;
}

int mda_baseboard_set_parameters(struct aslenum_device *baseboard, struct mda_device_params *params)
{
   int error = 0;

   if (baseboard->baseboard_type == board_type_FireBird){ /* Requires global space */
      if (params->memory_length == 0){
         error = -EINVAL;
         dev_err(baseboard->device, "%s: Zero global memory length specified for FireBird\n", __FUNCTION__);
         goto bail;
      }
      if (((size_t)params->memory_length + (size_t)params->memory_offset) > baseboard->board_address.kvsize){
         error = -EINVAL;
         dev_err(baseboard->device, "%s: memory area exceeds BAR bounds\n", __FUNCTION__);
         goto bail;
      }
      if (params->io_length){ /* FireBird does not support I/O access */
         error = -EINVAL;
         dev_err(baseboard->device, "%s: Unsupported global I/O length specified for FireBird\n", __FUNCTION__);
         goto bail;
      }
      baseboard->device_address.kvsize = (size_t)params->memory_length;
      baseboard->device_address.kvaddr = baseboard->board_address.kvaddr + params->memory_offset;
   }
   else if (baseboard->baseboard_type == board_type_Phoenix){
      if ((params->memory_length) || (params->io_length)){
         dev_err(baseboard->device, "%s: Unsupported global registers specified for Phoenix\n", __FUNCTION__);
         error = -EINVAL;
         goto bail;
      }
   }
   else{ /* Unrecognised board */
      dev_err(baseboard->device, "%s: Attempt to configure unrecognised board\n", __FUNCTION__);
      error = -ENODEV;
      goto bail;
   }

bail:
   return error;
}

void mda_child_release(struct device * device)
{
   struct mda_device *mda_device = to_mda_device(device);

   if (mda_device->id.driver_matching_name) kfree(mda_device->id.driver_matching_name);
   mda_channel_put(mda_device->parent);
   kfree(mda_device);
}

void mda_channel_release(struct device * device)
{
   struct mda_channel *mda_channel = to_mda_channel(device);

   mda_root_node_put(mda_channel->parent);
   kfree(mda_channel);
}

static ssize_t show_name(struct device *device, struct device_attribute *attr, char *buf);
static ssize_t show_modalias(struct device *device, struct device_attribute *attr, char *buf);
static ssize_t show_devname(struct device *device, struct device_attribute *attr, char *buf);

static DEVICE_ATTR(modalias, S_IRUGO, show_modalias, NULL);
static DEVICE_ATTR(name, S_IRUGO, show_name, NULL);
static DEVICE_ATTR(devname, S_IRUGO, show_devname, NULL);

static struct attribute *mda_channel_attrs[] = {
   &dev_attr_name.attr,
   &dev_attr_devname.attr,
   NULL
};

static struct attribute_group mda_channel_attr_group = {
   .attrs      = mda_channel_attrs,
};

static const struct attribute_group *mda_channel_attr_groups[] = {
   &mda_channel_attr_group,
   NULL
};

static struct device_type mda_channel_type = {
   .groups  = mda_channel_attr_groups,
   .uevent  = mda_channel_uevent,
   .release = mda_channel_release,
};

ssize_t show_device_matching(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%s\n", mda_device->id.driver_matching_name);
}

ssize_t show_device_io_length(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%08X\n", mda_device->params.io_length);
}

ssize_t show_device_io_offset(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%08X\n", mda_device->params.io_offset);
}

ssize_t show_device_memory_length(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%08X\n", mda_device->params.memory_length);
}

ssize_t show_device_memory_offset(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%08X\n", mda_device->params.memory_offset);
}

ssize_t show_device_interrupt_register(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%08X\n", mda_device->params.interrupt_register);
}

ssize_t show_device_interrupt_mask(struct device *device, struct device_attribute *attr, char *buf)
{
   struct mda_device *mda_device = to_mda_device(device);

   return sprintf(buf, "%08X\n", mda_device->params.interrupt_mask);
}

static DEVICE_ATTR(matching, S_IRUGO, show_device_matching, NULL);
static DEVICE_ATTR(io_length, S_IRUGO, show_device_io_length, NULL);
static DEVICE_ATTR(io_offset, S_IRUGO, show_device_io_offset, NULL);
static DEVICE_ATTR(memory_length, S_IRUGO, show_device_memory_length, NULL);
static DEVICE_ATTR(memory_offset, S_IRUGO, show_device_memory_offset, NULL);
static DEVICE_ATTR(interrupt_register, S_IRUGO, show_device_interrupt_register, NULL);
static DEVICE_ATTR(interrupt_mask, S_IRUGO, show_device_interrupt_mask, NULL);

static struct attribute *mda_child_attrs[] = {
   &dev_attr_name.attr,
   &dev_attr_devname.attr,
   &dev_attr_modalias.attr,
   &dev_attr_io_length.attr,
   &dev_attr_io_offset.attr,
   &dev_attr_memory_length.attr,
   &dev_attr_memory_offset.attr,
   &dev_attr_interrupt_register.attr,
   &dev_attr_interrupt_mask.attr,
   &dev_attr_matching.attr,
   NULL
};

static struct attribute_group mda_child_attr_group = {
   .attrs      = mda_child_attrs,
};

static const struct attribute_group *mda_child_attr_groups[] = {
   &mda_child_attr_group,
   NULL
};

static struct device_type mda_child_type = {
   .groups  = mda_child_attr_groups,
   .uevent  = mda_child_uevent,
   .release = mda_child_release,
};

static struct mda_device * mda_verify_child_device(struct device * device)
{
   if (device->type == &mda_child_type) return to_mda_device(device);
   else return NULL;
}

ssize_t show_devname(struct device * device, struct device_attribute * attr, char *buf)
{
   return sprintf(buf, "%s\n", dev_name(device));
}

ssize_t show_modalias(struct device * device, struct device_attribute * attr, char *buf)
{
   struct mda_device *mda_dev = to_mda_device(device);

   return sprintf(buf, "MODALIAS=%s:%08X:%02u:%02u:%02u:%02u\n",
                        MDA_MODULE_PREFIX,
                        (unsigned int)mda_dev->id.device_name_type,
                        mda_dev->parent->parent->parent->minor,
                        mda_dev->parent->parent->instance,
                        mda_dev->parent->instance,
                        mda_dev->instance);
}

ssize_t show_root_node_boardtype(struct device * device, struct device_attribute * attr, char *buf)
{
   struct mda_root_node *mda_root_node = to_mda_root_node(device);
   char *szDevice = NULL;

   switch(mda_root_node->parent->pcidev->device){
      case FBD_V6_2PE8_DEVICE_ID:
      case FBD_A7_2PE4_DEVICE_ID:
      case FBD_KU_3PE4_DEVICE_ID:
      case FBD_KU_3PE8_DEVICE_ID:
         szDevice = ASLENUM_STRING_FIREBIRD;
         break;
      case PHX_S3_1PEX_DEVICE_ID:
         szDevice = ASLENUM_STRING_PHOENIX;
         break;
      default:
         szDevice = "Unknown";
   }
   return sprintf(buf, "%s\n", szDevice);
}

ssize_t show_root_node_device(struct device * device, struct device_attribute * attr, char *buf)
{
   struct mda_root_node *mda_root_node = to_mda_root_node(device);

   return sprintf(buf, "%04X\n", mda_root_node->parent->pcidev->subsystem_device);
}

static DEVICE_ATTR(boardtype, S_IRUGO, show_root_node_boardtype, NULL);
static DEVICE_ATTR(device, S_IRUGO, show_root_node_device, NULL);

static struct attribute *mda_root_node_attrs[] = {
   &dev_attr_name.attr,
   &dev_attr_boardtype.attr,
   &dev_attr_device.attr,
   NULL
};

static struct attribute_group mda_root_node_attr_group = {
   .attrs      = mda_root_node_attrs,
};

static const struct attribute_group *mda_root_node_attr_groups[] = {
   &mda_root_node_attr_group,
   NULL
};

void mda_root_node_release(struct device * device)
{
   struct mda_root_node *mda_root_node = to_mda_root_node(device);

   mda_baseboard_put(mda_root_node->parent);
   kfree(mda_root_node);
}

struct device_type mda_root_node_type = {
   .groups      = mda_root_node_attr_groups,
   .release   = mda_root_node_release,
};

struct mda_root_node * mda_root_node_create(struct aslenum_device *baseboard, int index)
{
   int error = 0;
   struct mda_root_node *mda_root_node = NULL;

   baseboard = mda_baseboard_get(baseboard);
   if (!baseboard){
      error = -ENOENT;
      goto bail;
   }

   mda_root_node = kzalloc(sizeof(*mda_root_node), GFP_KERNEL);
   if (!mda_root_node){
      dev_err(baseboard->device, "%s: Allocation failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto bail;
   }
   INIT_LIST_HEAD(&mda_root_node->this_mda_root_node);
   INIT_LIST_HEAD(&mda_root_node->channel_list);
   INIT_LIST_HEAD(&mda_root_node->irq_register_list);
   mda_root_node->parent = baseboard;
   mda_root_node->instance = index;
   spin_lock_init(&mda_root_node->lock);
   spin_lock_init(&mda_root_node->irq_lock);
   sprintf(mda_root_node->name,"%s%02u:%02u", MDA_MODULE_PREFIX,
           mda_root_node->parent->minor,
           mda_root_node->instance);

   dev_set_name ( &mda_root_node->device, mda_root_node->name );
   mda_root_node->device.bus = &mda_bus_type;
   mda_root_node->device.type = &mda_root_node_type;
   mda_root_node->device.parent = &baseboard->pcidev->dev;

   error = device_register( &mda_root_node->device );
   if (error){
      dev_err(baseboard->device, "%s: device register failed with %d\n", __FUNCTION__, error);
      put_device(&mda_root_node->device);
      mda_root_node = NULL;
      goto bail;
   }
   spin_lock(&baseboard->mda_root_node_lock);
   mda_root_node = mda_root_node_get(mda_root_node);
   if (mda_root_node) {
      list_add_tail(&mda_root_node->this_mda_root_node, &baseboard->mda_root_node_list);
   }
   spin_unlock(&baseboard->mda_root_node_lock);

bail:
   if (!mda_root_node && baseboard) mda_baseboard_put(baseboard);
   return mda_root_node;
}

struct mda_channel * mda_channel_create(struct mda_root_node *root_node, int index)
{
   int error = 0;
   struct mda_channel *mda_channel = NULL;

   root_node = mda_root_node_get(root_node);
   if (!root_node){
      error = -ENOENT;
      goto bail;
   }

   mda_channel = kzalloc(sizeof(*mda_channel), GFP_KERNEL);
   if (!mda_channel){
      dev_err(&root_node->device, "%s: Allocation failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto bail;
   }
   INIT_LIST_HEAD(&mda_channel->this_mda_channel);
   INIT_LIST_HEAD(&mda_channel->device_list);
   mda_channel->parent = root_node;
   mda_channel->instance = index;
   spin_lock_init(&mda_channel->lock);
   sprintf(mda_channel->name,"%02u:%02u:%02u",
           root_node->parent->minor,
           root_node->instance, index);

   dev_set_name( &mda_channel->device, mda_channel->name );
   mda_channel->device.bus = &mda_bus_type;
   mda_channel->device.type = &mda_channel_type;
   mda_channel->device.parent = &root_node->device;

   error = device_register(&mda_channel->device);
   if (error){
      dev_err(&root_node->device, "%s: device register failed with %d\n", __FUNCTION__, error);
      put_device(&mda_channel->device);
      mda_channel = NULL;
      goto bail;
   }
   spin_lock(&root_node->lock);
   mda_channel = mda_channel_get(mda_channel);
   if (mda_channel) {
      list_add_tail(&mda_channel->this_mda_channel, &root_node->channel_list);
   }
   spin_unlock(&root_node->lock);

bail:
   if (!mda_channel && root_node) mda_root_node_put(root_node);
   return mda_channel;
}

struct mda_device * mda_device_create(struct mda_channel *channel,
                                      struct mda_device_id *id,
                                      struct mda_device_params *params)
{
   struct mda_device * mda_device = NULL;
   int error = 0;
   char *kvaddr = NULL;

   channel = mda_channel_get(channel);
   if (!channel){
      error = -ENOENT;
      goto bail;
   }

   mda_device = kzalloc(sizeof(*mda_device), GFP_KERNEL);
   if (!mda_device){
      dev_err(&channel->device, "%s: Allocation failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto bail;
   }
   mda_device->id.driver_matching_name = kzalloc(strlen(id->driver_matching_name) + strlen("\0"), GFP_KERNEL);
   if (mda_device->id.driver_matching_name == NULL){
      kfree(mda_device);
      mda_device = NULL;
      error = -ENOMEM;
      goto bail;
   }
   strcpy(mda_device->id.driver_matching_name, id->driver_matching_name);
   mda_device->id.device_name_type = id->device_name_type;
   INIT_LIST_HEAD(&mda_device->this_mda_device);
   mda_device->parent = channel;
   mda_device->params = *params;
   mda_device->channel = channel->instance;

   if (params->memory_length){ /* Memory mapped device */
      kvaddr = channel->parent->parent->board_address.kvaddr;
      mda_device->device_address.kvaddr =  kvaddr + params->memory_offset;
      mda_device->device_address.kvsize = (size_t)params->memory_length;
   }
   else if (params->io_length){ /* I/O mapped device */
      kvaddr = channel->parent->parent->ioport_address.kvaddr;
      mda_device->device_address.kvaddr = kvaddr + params->io_offset;
      mda_device->device_address.kvsize = (size_t)params->io_length;
   }
   else{ /* Pseudo device with no physical registers */
      mda_device->device_address.kvaddr = NULL;
      mda_device->device_address.kvsize = 0;
   }

   spin_lock(&channel->lock);
   mda_device->instance = channel->device_count++;
   spin_unlock(&channel->lock);

   sprintf(mda_device->name, "%02u:%02u:%02u:%02u",
                             channel->parent->parent->minor,
                             channel->parent->instance,
                             channel->instance,
                             mda_device->instance);

   dev_set_name(&mda_device->device, mda_device->name);
   mda_device->device.bus = &mda_bus_type;
   mda_device->device.type = &mda_child_type;
   mda_device->device.parent = &channel->device;

   error = device_register(&mda_device->device);
   if (error){
      dev_err(&channel->device, "%s: device register failed with %d\n", __FUNCTION__, error);
      put_device(&mda_device->device);
      mda_device = NULL;
      goto bail;
   }

   spin_lock(&channel->lock);
   mda_device = mda_dev_get(mda_device);
   if (mda_device) {
      list_add_tail(&mda_device->this_mda_device, &channel->device_list);
   }
   spin_unlock(&channel->lock);

   if (!mda_device) goto bail;

   dev_printk(KERN_INFO,
              &channel->device,
              "%s created %s@%s(%08X/%08X),INT(%08X/%08X)\n",
              __FUNCTION__,
              mda_device->id.driver_matching_name,
              mda_device->params.memory_length ? "MM" : "IO",
              mda_device->params.memory_length ? mda_device->params.memory_offset : mda_device->params.io_offset,
              mda_device->params.memory_length ? mda_device->params.memory_length : mda_device->params.io_length,
              mda_device->params.interrupt_register,
              mda_device->params.interrupt_mask);

bail:
   if (!mda_device && channel) mda_channel_put(channel);
   return mda_device;
}

int mda_create_nodes(struct aslenum_device *baseboard)
{
   char * pDescription = NULL;
   uint32_t config_vendor     = 0;
   uint32_t config_device     = 0;
   uint32_t config_sub_vendor = 0;
   uint32_t config_sub_device = 0;
   uint32_t config_revision   = 0; 
   uint32_t vendor     = baseboard->pcidev->vendor;
   uint32_t device     = baseboard->pcidev->device;
   uint32_t sub_vendor = baseboard->pcidev->subsystem_vendor;
   uint32_t sub_device = baseboard->pcidev->subsystem_device;
   uint32_t revision   = baseboard->pcidev->revision;
   uint32_t index = 0;

   aslenum_get_fbd_config_desc( index, &pDescription );
   while(pDescription){
      aslenum_get_fbd_config_vendor    ( index, &config_vendor );
      aslenum_get_fbd_config_device    ( index, &config_device );
      aslenum_get_fbd_config_sub_vendor( index, &config_sub_vendor );
      aslenum_get_fbd_config_sub_device( index, &config_sub_device );
      aslenum_get_fbd_config_revision  ( index, &config_revision ); 
      if ((config_vendor == vendor) && (config_sub_vendor == sub_vendor) &&
          (config_device == device) && (config_sub_device == sub_device) && (config_revision == revision)){
         dev_printk(KERN_INFO, baseboard->device, "Detected Active Silicon %s\n", pDescription);
         return mda_config_board(index, baseboard);
      }
      index++;
      aslenum_get_fbd_config_desc( index, &pDescription );
   }
   index = 0;
   aslenum_get_fbd_config_desc( index, &pDescription ); 
   while(pDescription){
      aslenum_get_fbd_config_vendor    ( index, &config_vendor );
      aslenum_get_fbd_config_device    ( index, &config_device );
      aslenum_get_fbd_config_sub_vendor( index, &config_sub_vendor );
      aslenum_get_fbd_config_sub_device( index, &config_sub_device );
      aslenum_get_fbd_config_revision  ( index, &config_revision ); 
      if ((config_vendor == vendor) && (config_sub_vendor == sub_vendor) &&
          (config_device == device) && (config_sub_device == sub_device) && (config_revision == 0)){
         dev_printk(KERN_INFO, baseboard->device, "Detected Active Silicon %s\n", pDescription);
         return mda_config_board(index, baseboard);
      }
      index++;
      aslenum_get_fbd_config_desc( index, &pDescription ); 
   }
   return -EIO;
}

void mda_root_node_destroy(struct mda_root_node *mda_root_node)
{
   if (!mda_root_node) goto bail;
   mda_destroy_channels(mda_root_node);
   device_unregister(&mda_root_node->device);
   mda_root_node_put(mda_root_node);
bail:
   return;
}

void mda_channel_destroy(struct mda_channel *channel)
{
   if (!channel) goto bail;
   mda_destroy_devices(channel);
   device_unregister(&channel->device);
   mda_channel_put(channel);
bail:
   return;
}

void mda_device_destroy(struct mda_device *mda_device)
{
   if (!mda_device) goto bail;
   if (mda_device->driver){
      if (mda_device->driver->remove){
         mda_device->driver->remove(mda_device);
      }
   }
   device_unregister(&mda_device->device);
   mda_dev_put(mda_device);

bail:
   return;
}

void mda_destroy_channels(struct mda_root_node *root_node)
{
   struct mda_channel *channel = NULL;
   struct mda_channel *temp = NULL;

   spin_lock(&root_node->lock);
   list_for_each_entry_safe(channel, temp, &root_node->channel_list, this_mda_channel){
      list_del(&channel->this_mda_channel);
      mda_channel_destroy(channel);
   }
   spin_unlock(&root_node->lock);
}

void mda_destroy_devices(struct mda_channel *channel)
{
   struct mda_device *mda_device = NULL;
   struct mda_device *temp = NULL;

   spin_lock(&channel->lock);
   list_for_each_entry_safe(mda_device, temp, &channel->device_list, this_mda_device){
      list_del(&mda_device->this_mda_device);
      mda_device_destroy(mda_device);
   }
   spin_unlock(&channel->lock);
}


void mda_destroy_nodes(struct aslenum_device *baseboard)
{
   struct mda_root_node * mda_root_node = NULL;
   struct mda_root_node * temp = NULL;

   spin_lock(&baseboard->mda_root_node_lock);
   list_for_each_entry_safe(mda_root_node, temp, &baseboard->mda_root_node_list, this_mda_root_node){
      list_del(&mda_root_node->this_mda_root_node);
      mda_root_node_destroy(mda_root_node);
   }
   spin_unlock(&baseboard->mda_root_node_lock);
   return;
}

ssize_t show_name(struct device * device, struct device_attribute * attr, char *buf)
{
   ssize_t bytes_out = 0;

   if (!device) return -ENODEV;

   if (device->type == &mda_child_type){ /* child device */
      bytes_out =  sprintf(buf, "%s\n", to_mda_device(device)->name);
   }
   else if (device->type == &mda_channel_type){
     bytes_out =  sprintf(buf, "%s\n", to_mda_channel(device)->name);
   }
   else if (device->type == &mda_root_node_type){ /* mda_root_node device */
      bytes_out =  sprintf(buf, "%s\n", to_mda_root_node(device)->name);
   }
   else{
      dev_err(device, "%s: Unrecognised device->type", __FUNCTION__);
      bytes_out =  sprintf(buf, "%s:MDA UNRECOGNISED\n", __FUNCTION__);
   }
   return bytes_out;
}


#if defined CONFIG_HOTPLUG
static int mda_bus_uevent(struct device *device, struct kobj_uevent_env *env)
{
   struct mda_root_node * mda_root_node = NULL;

   if (!device) return -ENODEV;


   if (device->type == &mda_root_node_type){
      mda_root_node = to_mda_root_node(device);
   }
   else if (device->type == &mda_channel_type){
      mda_root_node = to_mda_channel(device)->parent;
   }
   else if (device->type == &mda_child_type){
      mda_root_node = to_mda_device(device)->parent->parent;
   }
   else{
      dev_err(device, "%s: unrecognised device->type\n", __FUNCTION__);
   }

   if (!mda_root_node) return -ENODEV;

   if (add_uevent_var(env, "MDA_BUS=%u", mda_root_node->parent->minor)) return -ENOMEM;
   if (add_uevent_var(env, "MDA_ROOTNODE=%u", mda_root_node->instance)) return -ENOMEM;
   if (add_uevent_var(env, "MDA_PARENT_DEVICE=/dev/%s", dev_name(mda_root_node->parent->device))) return -ENOMEM;

   return 0;
}

int mda_child_uevent(struct device * device, struct kobj_uevent_env * env)
{
   struct mda_device *mda_dev = to_mda_device(device);

   if (!mda_dev) return -ENODEV;

   switch(mda_dev->id.device_name_type){
   case MDAENUM_DEVICE_NAME_REQUEST_TYPE_DMA:
      if (add_uevent_var(env, "MODALIAS=%s", "asldma")) return -ENOMEM;
      break;
   case MDAENUM_DEVICE_NAME_REQUEST_TYPE_SERIAL:
      if (add_uevent_var(env, "MODALIAS=%s", "aslser")) return -ENOMEM;
      break;
   case MDAENUM_DEVICE_NAME_REQUEST_TYPE_CXP:
      if (add_uevent_var(env, "MODALIAS=%s", "aslcxp")) return -ENOMEM;
      break;
   default:
      dev_printk(KERN_WARNING, device, "%s: Unrecognised device type %u\n", __FUNCTION__, (unsigned int)mda_dev->id.device_name_type);
      break;
   }
   if (add_uevent_var(env, "MDA_CHANNEL=%u", mda_dev->parent->instance)) return -ENOMEM;
   if (add_uevent_var(env, "MDA_CHILD=%u", mda_dev->instance)) return -ENOMEM;
   if (add_uevent_var(env, "MDA_PARENT=%s", dev_name(&mda_dev->parent->device))) return -ENOMEM;

   return 0;

}

static int mda_channel_uevent(struct device *device, struct kobj_uevent_env *env)
{
   struct mda_channel *mda_channel = to_mda_channel(device);

   if (!mda_channel) return -ENODEV;

   if (add_uevent_var(env, "MDA_CHANNEL=%u", mda_channel->instance)) return -ENOMEM;
   if (add_uevent_var(env, "MDA_PARENT=%s", dev_name(&mda_channel->parent->device))) return -ENOMEM;

   return 0;
}

#else
static int mda_bus_uevent(struct device *dev, struct kobj_uevent_env *env)
{
   return -ENODEV;
}

static int mda_child_uevent(struct device *device, struct kobj_uevent_env *env)
{
   return -ENODEV;
}
static int mda_channel_uevent(struct device *device, struct kobj_uevent_env *env)
{
   return -ENODEV;
}
#endif

enum baseboard_type mda_get_baseboard_type(struct mda_device *mda_device)
{
   return mda_device->parent->parent->parent->baseboard_type;
}


struct aslenum_device * device_to_baseboard(struct device *device)
{
   struct aslenum_device *baseboard = NULL;

   if (device->type == &mda_root_node_type){
      baseboard = to_mda_root_node(device)->parent;
   }
   else if (device->type == &mda_channel_type){
      baseboard = to_mda_channel(device)->parent->parent;
   }
   else if (device->type == &mda_child_type){
      baseboard = to_mda_device(device)->parent->parent->parent;
   }
   return baseboard;
}

int __must_check mda_do_register_driver(struct mda_driver *mda_driver, struct module *owner, const char *mod_name)
{
   int error = 0;

   mda_driver->driver.name = mda_driver->name;
   mda_driver->driver.bus = &mda_bus_type;
   mda_driver->driver.owner = owner;
   mda_driver->driver.mod_name = mod_name;
   INIT_LIST_HEAD(&mda_driver->this_mda_driver);
   error = driver_register(&mda_driver->driver);
   if (error){
      printk(KERN_ERR "%s: driver_register returned %d for %s %s\n",
                      __FUNCTION__,
                      error,
                      mod_name ? mod_name : "Unknown modname",
                      mda_driver->name ? mda_driver->name : "Unknown driver name");
   }
   return error;
}

void mda_unregister_driver(struct mda_driver *mda_driver)
{
   printk(KERN_INFO "%s: unregistering %s %s\n",
                   __FUNCTION__,
                   mda_driver->driver.mod_name ? mda_driver->driver.mod_name : "Unknown modname",
                   mda_driver->driver.name ?  mda_driver->driver.name : "Unknown driver name");
   driver_unregister(&mda_driver->driver);
   return;
}

EXPORT_SYMBOL(mda_do_register_driver);
EXPORT_SYMBOL(mda_unregister_driver);
EXPORT_SYMBOL(mda_bus_type);
EXPORT_SYMBOL(mda_set_drvdata);
EXPORT_SYMBOL(mda_get_drvdata);
EXPORT_SYMBOL(mda_root_node_type);
EXPORT_SYMBOL(device_to_baseboard);
EXPORT_SYMBOL(mda_get_baseboard_type);

