#include "precomp.h"
#include "boards.h"

static const uint32_t GLOBAL_MEM_LENGTH                      = 0x00080000;
static const uint32_t GLOBAL_MEM_OFFSET                      = 0x00000000;

static const uint32_t ACQ_IRQ_REG                            = 0x00001000;
static const uint32_t SER_IRQ_REG                            = 0x00001004;
static const uint32_t CXP_IRQ_REG                            = 0x00001008;
static const uint32_t GIO_IRQ_REG                            = 0x0000100C;

static const uint32_t CLACQ_MEM_LENGTH                       = 0x00140000;
static const uint32_t CXPACQ_MEM_LENGTH                      = 0x00740000;
static const uint32_t CXPCOMMS_MEM_LENGTH                    = 0x00002000;
static const uint32_t CLSER_MEM_LENGTH                       = 0x00010000;
static const uint32_t GIO_MEM_LENGTH                         = 0x00002000;
static const uint32_t PHXSER_MEM_LENGTH                      = 0x00001000;

static const uint32_t CHAN_MEM_OFFSET                        = 0x00200000;

static const uint32_t PHXSER_MIR_CHANNEL_1                   = 0x00380000;
static const uint32_t PHXSER_MIR_CHANNEL_2                   = 0x00381000;
static const uint32_t PHXSER_MIR_CHANNEL_3                   = 0x00382000;
static const uint32_t PHXSER_MIR_CHANNEL_4                   = 0x00383000;

static const uint32_t ACQ_VISIBLE                            = 0x00000000;
static const uint32_t CXP_VISIBLE                            = 0x00000000;
static const uint32_t SER_VISIBLE                            = 0x00000001;

static const uint32_t PHX_MEM_OFFSET                         = 0x00000000;
static const uint32_t PHX_MEM_LENGTH                         = 0x01000000;
static const uint32_t PHX_SER_OFFSET_A                       = 0x00000000;
static const uint32_t PHX_SER_OFFSET_B                       = 0x00000008;
static const uint32_t PHX_SER_LENGTH                         = 0x00000008;

static struct mda_root_node * mda_config_create_root(struct aslenum_device *baseboard)
{
   struct mda_root_node *root_node = NULL;
   struct mda_device_params params = {0};

   params.memory_offset = GLOBAL_MEM_OFFSET;
   params.memory_length = GLOBAL_MEM_LENGTH;
   if (mda_baseboard_set_parameters(baseboard, &params)) goto bail;

   root_node = mda_root_node_create(baseboard, 0);

bail:
   return root_node;
}

static struct mda_device * mda_dma_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask)
{
   struct mda_device_params params = {0};
   struct mda_device_id id = {0};

   id.driver_matching_name   = "asldma";
   id.device_name_type       = MDAENUM_DEVICE_NAME_REQUEST_TYPE_DMA;
   params.interrupt_mask     = interrupt_mask;
   params.interrupt_register = interrupt_register;
   params.memory_offset      = memory_offset;
   params.memory_length      = memory_length;

   return mda_device_create(parent_channel, &id, &params);
}

static struct mda_device * mda_ser_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask,
                                                 const uint32_t      FullDuplex)
{
   struct mda_device_params params = {0};
   struct mda_device_id id = {0};

   id.driver_matching_name   = "aslser";
   id.device_name_type       = MDAENUM_DEVICE_NAME_REQUEST_TYPE_SERIAL;
   params.interrupt_mask     = interrupt_mask;
   params.interrupt_register = interrupt_register;
   params.memory_offset      = memory_offset;
   params.memory_length      = memory_length;

   return mda_device_create(parent_channel, &id, &params);
}

static struct mda_device * mda_cls_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask)
{
   return mda_ser_device_create(parent_channel, memory_offset, memory_length, interrupt_register, interrupt_mask, 1);
}

static struct mda_device * mda_232_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask)
{
   return mda_ser_device_create(parent_channel, memory_offset, memory_length, interrupt_register, interrupt_mask, 1);
}

static struct mda_device * mda_485_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask)
{
   return mda_ser_device_create(parent_channel, memory_offset, memory_length, interrupt_register, interrupt_mask, 0);
}

static struct mda_device * mda_cxp_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask)
{
   struct mda_device_params params = {0};
   struct mda_device_id id = {0};

   id.driver_matching_name   = "aslcxp";
   id.device_name_type       = MDAENUM_DEVICE_NAME_REQUEST_TYPE_CXP;
   params.interrupt_mask     = interrupt_mask;
   params.interrupt_register = interrupt_register;
   params.memory_offset      = memory_offset;
   params.memory_length      = memory_length;

   return mda_device_create(parent_channel, &id, &params);
}

static struct mda_device * mda_gio_device_create(struct mda_channel *parent_channel,
                                                 const uint32_t      memory_offset,
                                                 const uint32_t      memory_length,
                                                 const uint32_t      interrupt_register,
                                                 const uint32_t      interrupt_mask)
{
   struct mda_device_params params = {0};
   struct mda_device_id id = {0};

   id.driver_matching_name   = "aslgio";                             /* 999 */
   id.device_name_type       = MDAENUM_DEVICE_NAME_REQUEST_TYPE_GIO; /* 999 */
   params.interrupt_mask     = interrupt_mask;
   params.interrupt_register = interrupt_register;
   params.memory_offset      = memory_offset;
   params.memory_length      = memory_length;

   return mda_device_create(parent_channel, &id, &params);
}

static int mda_config_create_channels(struct mda_root_node *root_node, struct mda_channel **mda_channels, size_t nChannels)
{
   int error = -EIO;
   uint32_t channel_index = 0;

   for (channel_index = 0; channel_index < (uint32_t) nChannels; channel_index++)
   {
      mda_channels[channel_index] = mda_channel_create(root_node, channel_index);
      if ( NULL == mda_channels[channel_index] ) goto bail;
   }
   error = 0;
bail:
   return error;
}

int mda_config_board(uint32_t index, tAslEnumDevice *baseboard)
{
   tMdaRootNode *root_node = NULL;
   tMdaChannel **mda_channels = NULL;
   tMdaChannel *channel = NULL;
   uint32_t i = 0;
   uint32_t num_channels = 0;
   uint32_t num_dma_channels = 0;
   uint32_t num_cls_channels = 0; 
   uint32_t num_cxp_channels = 0;
   uint32_t num_gio_channels = 0;
   uint32_t num_485_channels = 0;
   uint32_t num_232_channels = 0;
   uint32_t acq_mem_length = CLACQ_MEM_LENGTH;
   uint32_t current_channel = 0;
   uint32_t mem_offset = 0;
   uint32_t irq_mask = 0;

   root_node = mda_config_create_root(baseboard);
   if (!root_node) return MDA_CONFIG_ERROR;
   aslenum_get_num_channels( index, &num_channels );
   mda_channels = kzalloc( sizeof(*mda_channels) * num_channels, GFP_KERNEL );
   if (mda_channels == NULL) return MDA_CONFIG_ERROR;
   if (mda_config_create_channels(root_node, mda_channels, num_channels)) return MDA_CONFIG_ERROR;

   aslenum_get_num_cxp_channels( index, &num_cxp_channels );
   if( num_cxp_channels )
      acq_mem_length = CXPACQ_MEM_LENGTH;
   aslenum_get_num_dma_channels( index, &num_dma_channels );
   for( i = 0; i < num_dma_channels; i++ )
   {
      channel = mda_channels[i];
      aslenum_get_acq_info( index, i, &mem_offset, &irq_mask );
	  if (!mda_dma_device_create(channel, mem_offset, acq_mem_length, ACQ_IRQ_REG, irq_mask)) return MDA_CONFIG_ERROR;
   }
   current_channel += num_dma_channels;

   aslenum_get_num_cls_channels( index, &num_cls_channels );
   if( num_cls_channels )
   {
      /* The cls devices are created with the same channels as the dma devices.
 */
      for( i = 0; i < num_cls_channels; i++ )
{
         channel = mda_channels[i];
         aslenum_get_cls_info( index, i, &mem_offset, &irq_mask );
         if (!mda_cls_device_create(channel, mem_offset, CLSER_MEM_LENGTH, SER_IRQ_REG, irq_mask)) return MDA_CONFIG_ERROR;
      }
   }
   else if( num_cxp_channels )
   {
      /* The cxp devices are created with the same channels as the dma devices,
       */
      for( i = 0; i < num_dma_channels; i++ )
      {
         channel = mda_channels[i];
         aslenum_get_cxp_info( index, i, &mem_offset, &irq_mask );
         if (!mda_cxp_device_create(channel, mem_offset, CXPCOMMS_MEM_LENGTH, CXP_IRQ_REG, irq_mask)) return MDA_CONFIG_ERROR; 
      }
   }
   aslenum_get_num_gio_channels( index, &num_gio_channels );
   if( num_gio_channels )
   {
      /* The gio devices are created with the same channels as the dma devices,
       */
      for( i = 0; i < num_gio_channels; i++ )
      {
         channel = mda_channels[i];
         aslenum_get_gio_info( index, i, &mem_offset, &irq_mask );
         if (!mda_gio_device_create(channel, mem_offset, GIO_MEM_LENGTH, GIO_IRQ_REG, irq_mask)) return MDA_CONFIG_ERROR; 
      }
   }
   /* The 485 and 232 devices are created using subsequent channels.
    */
   aslenum_get_num_485_channels( index, &num_485_channels );
   for( i = current_channel; i < current_channel + num_485_channels; i++ )
   {
      channel = mda_channels[i]; 
      aslenum_get_485_info( index, i - current_channel, &mem_offset, &irq_mask );
      if (!mda_485_device_create(channel, mem_offset, CLSER_MEM_LENGTH, SER_IRQ_REG, irq_mask)) return MDA_CONFIG_ERROR;
   }
   current_channel += num_485_channels;
   aslenum_get_num_232_channels( index, &num_232_channels );
   for( i = current_channel; i < current_channel + num_232_channels; i++ )
   {
      channel = mda_channels[i];
      aslenum_get_232_info( index, i - current_channel, &mem_offset, &irq_mask );
      if (!mda_232_device_create(channel, mem_offset, CLSER_MEM_LENGTH, SER_IRQ_REG, irq_mask)) return MDA_CONFIG_ERROR;
   }

   return MDA_CONFIG_OK;
}
