#include "precomp.h"

static int debug_reglist = 0;

static int mda_operation_matches(uint32_t operation, uint32_t comparand)
{
   return((operation & comparand) == comparand);
}

static long mda_ioctl_process_element(struct device *device,
                                      struct aslenum_device *baseboard,
                                      uint32_t *element,
                                      struct mda_io_address *io_address,
                                      int writeok)
{
   long error = 0;
   uint32_t operation = element[0];
   uint32_t data = element[2];

   if (debug_reglist) dev_printk(KERN_INFO, device, "%s(ENTRY): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);

   if(mda_operation_matches(operation, CDA_REG_WRITE)){
      char *address   = io_address->kvaddr + element[1];
      if (!writeok){
         dev_err(device, "%s: Illegal write operation\n", __FUNCTION__);
         error = -EPERM;
         goto bail;
      }
      else if (mda_operation_matches(operation, CDA_REG_WRITE_32BITS)){
         __raw_writel(data, address);
      }
      else if (mda_operation_matches(operation, CDA_REG_WRITE_16BITS)){
         __raw_writew((uint16_t)data, address);
      }
      else if (mda_operation_matches(operation, CDA_REG_WRITE_8BITS)){
         __raw_writeb((uint8_t)data, address);
      }
      else{
         dev_err(device, "%s(REG_WRITE): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
         error = -EINVAL;
      }
   }
   else if (mda_operation_matches(operation, CDA_REG_READ)){
      char *address   = io_address->kvaddr + element[1];
      if (mda_operation_matches(operation, CDA_REG_READ_32BITS)){
         data = __raw_readl(address);
      }
      else if (mda_operation_matches(operation, CDA_REG_READ_16BITS)){
         data = (uint32_t)__raw_readw(address);
      }
      else if (mda_operation_matches(operation, CDA_REG_READ_8BITS)){
         data = (uint32_t)__raw_readb(address);
      }
       else{
         dev_err(device, "%s(REG_READ): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
         error = -EINVAL;
      }
   }
   else if (mda_operation_matches(operation, CDA_CNF_READ)){
     uint32_t address = element[1];
     if (mda_operation_matches(operation, CDA_CNF_READ_32BITS)){
        error = pci_read_config_dword(baseboard->pcidev, address, &data);
        if (error) dev_err(device, "%s(CNF_READ32): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
     }
     else if (mda_operation_matches(operation, CDA_CNF_READ_16BITS)){
        uint16_t tmp = 0;
        error = pci_read_config_word(baseboard->pcidev, address, &tmp);
        if (error) dev_err(device, "%s(CNF_READ16): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
        data = (uint32_t)tmp;
     }
     else if (mda_operation_matches(operation, CDA_CNF_READ_8BITS)){
        uint8_t tmp = 0;
        error = pci_read_config_byte(baseboard->pcidev, address, &tmp);
        if (error) dev_err(device, "%s(CNF_READ8): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
        data = (uint32_t)tmp;
     }
     else{
         dev_err(device, "%s(CNF_READ): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
         error = -EINVAL;
      }
   }
   else if (mda_operation_matches(operation, CDA_CNF_WRITE)){
     uint32_t address = element[1];
     if (mda_operation_matches(operation, CDA_CNF_WRITE_32BITS)){
        error = pci_write_config_dword(baseboard->pcidev, address, data);
        if (error) dev_err(device, "%s(CNF_WRITE32): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
     }
     else if (mda_operation_matches(operation, CDA_CNF_WRITE_16BITS)){
        error = pci_write_config_word(baseboard->pcidev, address, (uint16_t)data);
        if (error) dev_err(device, "%s(CNF_WRITE16): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
     }
     else if (mda_operation_matches(operation, CDA_CNF_WRITE_8BITS)){
        error = pci_write_config_byte(baseboard->pcidev, address, (uint8_t)data);
        if (error) dev_err(device, "%s(CNF_WRITE8): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
     }
     else{
         dev_err(device, "%s(CNF_WRITE): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
         error = -EINVAL;
      }
   }
   else{
      dev_err(device, "%s(CNF_UNKNOWN): entry %08x %08x %08x, base %p\n", __FUNCTION__, element[0], element[1], element[2], io_address->kvaddr);
      error = -EINVAL;
   }
bail:
   if (!error) element[2] = data;
   if (debug_reglist) dev_printk(KERN_INFO, device, "%s(EXIT): exit %08x %08x %08x\n", __FUNCTION__, element[0], element[1], element[2]);
   return error;
}

long mda_do_ioctl_reglist(struct device *device,
                          struct aslenum_device *baseboard,
                          struct mda_io_address *io_address,
                          mda_io_buffer *user_io,
                          int writeok)
{
   long error = 0;
   uint32_t *reglist = NULL;
   uint32_t element_total = 0;
   uint32_t element_index = 0;
   const uint32_t element_size = 3;
   const uint32_t element_size_bytes = element_size * sizeof(uint32_t);

   if ((user_io->input_buffer.user_address == NULL) || (user_io->input_buffer.user_size == 0)){
      dev_err(device, "%s: Invalid input buffer\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }
   if ((user_io->output_buffer.user_address == NULL) || (user_io->output_buffer.user_size == 0)){
      dev_err(device, "%s: Invalid output buffer\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }
   if (user_io->output_buffer.user_size < user_io->input_buffer.user_size){
      dev_err(device, "%s: Output buffer too small\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0) || MDA_EL8_KERNEL == 1
   if (!access_ok((u64)user_io->input_buffer.user_address, user_io->input_buffer.user_size)){
#else
   if (!access_ok(VERIFY_READ, (u64)user_io->input_buffer.user_address, user_io->input_buffer.user_size)){
#endif
      dev_err(device, "%s: Input buffer cannot be read\n", __FUNCTION__);
      error = -EACCES;
      goto bail;
   }
   element_total = user_io->input_buffer.user_size / element_size_bytes;
   if ((element_total * element_size_bytes) != user_io->input_buffer.user_size){
      dev_err(device, "%s: Input buffer size illegal\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }

   reglist = kzalloc(user_io->input_buffer.user_size, GFP_KERNEL);
   if (!reglist){
      dev_err(device, "%s: Buffer allocation failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto bail;
   }

   error = copy_from_user(reglist, user_io->input_buffer.user_address, user_io->input_buffer.user_size);
   if (error){
      dev_err(device, "%s: Buffer copy in (%ld bytes) failed with %ld\n", __FUNCTION__, user_io->input_buffer.user_size, error);
      goto bail;
   }

   if (debug_reglist) dev_printk(KERN_INFO, baseboard->device, "%s: %d elements\n", __FUNCTION__, element_total);

   for(element_index = 0; element_index < element_total; element_index++){
      error = mda_ioctl_process_element(device,
                                        baseboard,
                                        &reglist[element_index * element_size],
                                        io_address,
                                        writeok);
      if (error) goto bail;
   }

bail:
   if (!error){
      user_io->output_buffer.user_size = user_io->input_buffer.user_size;
      error = copy_to_user(user_io->output_buffer.user_address, reglist, user_io->output_buffer.user_size);
      if (error){
         dev_err(device, "%s: copy_to_user failed with %d\n", __FUNCTION__, (int)error);
         user_io->output_buffer.user_size = 0;
      }
   }
   if (reglist) kfree(reglist);
   return error;
}

#if 1
/* CLR999 !!! */
long aslenum_ioctl_reglist(struct aslenum_device *baseboard, struct file * file, mda_io_buffer *user_io)
{
   int writeok = 1;

   return mda_do_ioctl_reglist(baseboard->device, baseboard, &baseboard->device_address, user_io, writeok);
}
#else
long aslenum_ioctl_reglist(struct aslenum_device *baseboard, struct file * file, mda_io_buffer *user_io)
{
   int writeok = 0;

   if (aslenum_check_exclusive(baseboard, file) == 0) writeok = 1;
   return mda_do_ioctl_reglist(baseboard->device, baseboard, &baseboard->device_address, user_io, writeok);
}
#endif

static int aslenum_get_child_name(struct aslenum_device *baseboard,
	                          MDA_DeviceNameRequest *name_request,
	                          char *device_name,
	                          size_t *device_name_size)
{
   int error = -ENODEV;
   struct mda_root_node *root_node = NULL;
   size_t bytes_copied = 0;

   memset(device_name, 0, *device_name_size);

   spin_lock(&baseboard->mda_root_node_lock);
   list_for_each_entry(root_node, &baseboard->mda_root_node_list, this_mda_root_node){
      struct mda_channel *channel = NULL;

      spin_lock(&root_node->lock);
      list_for_each_entry(channel, &root_node->channel_list, this_mda_channel){

         spin_lock(&channel->lock);
         if (channel->instance == name_request->uiChannelIndex){
            uint32_t device_index = 0;
            struct mda_device *mda_device = NULL;
            list_for_each_entry(mda_device, &channel->device_list, this_mda_device){
               if (mda_device->id.device_name_type == name_request->uiType){
            		if (device_index++ == name_request->uiDeviceIndex){
            			bytes_copied = (size_t)snprintf(device_name, *device_name_size, "%s", mda_device->name) + 1;
            			bytes_copied += sizeof('\0'); /* Include terminator */
            			error = (bytes_copied <= *device_name_size) ? 0 : -EFAULT;
            		} /* end device index matches */
            	} /* end device type matches */
            } /* end for each device */
         } /* end channel matches */
         spin_unlock(&channel->lock);

      } /* end for each channel */
      spin_unlock(&root_node->lock);

   } /* end for each root_node */
   spin_unlock(&baseboard->mda_root_node_lock);

   *device_name_size = (error) ? 0 : bytes_copied;
   return error;
}


long aslenum_ioctl_get_child_name(struct aslenum_device *baseboard, mda_io_buffer *user_io)
{
   int error = -ENOTSUPP;
   MDA_DeviceNameRequest device_name_request = {0};
   char device_name[MDA_MAX_NAME_LENGTH] = {0};
   size_t device_name_length = MDA_MAX_NAME_LENGTH;

   if ((user_io->param_buffer.user_address == NULL) || (user_io->param_buffer.user_size < sizeof(device_name_request))){
      dev_err(baseboard->device, "%s: Parameter buffer too small\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }
   if ((user_io->output_buffer.user_address == NULL) || (user_io->output_buffer.user_size == 0)){
      dev_err(baseboard->device, "%s: Output buffer missing\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }

   error = copy_from_user(&device_name_request, user_io->param_buffer.user_address, sizeof(device_name_request));
   if (error){
      dev_err(baseboard->device, "%s: copy_from_user returned %d\n", __FUNCTION__, error);
      goto bail;
   }

   error = aslenum_get_child_name(baseboard, &device_name_request, device_name, &device_name_length);
   if (error){
      dev_err(baseboard->device, "%s: aslenum_get_child_name returned %d\n", __FUNCTION__, error);
      goto bail;
   }

   if (device_name_length > user_io->output_buffer.user_size){
	   dev_err(baseboard->device, "%s: output buffer too small\n", __FUNCTION__);
	   error = -EFAULT;
	   goto bail;
   }
   error = copy_to_user(user_io->output_buffer.user_address, device_name, device_name_length);
   if (error){
	   dev_err(baseboard->device, "%s: copy_to_user returned %d\n", __FUNCTION__, error);
	   goto bail;
   }

bail:
   return (long)error;
}

long aslenum_ioctl_set_lock_timeout(struct aslenum_file *pEnumFile, mda_io_buffer *user_io)
{
   long error = -EIO;

   if ((user_io->input_buffer.user_address == NULL) || (user_io->input_buffer.user_size < sizeof(pEnumFile->write_lock_timeout_ms))){
      dev_err(pEnumFile->pEnumDevice->device, "%s: Parameter buffer too small\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }

   error = copy_from_user(&pEnumFile->write_lock_timeout_ms, user_io->input_buffer.user_address, sizeof(pEnumFile->write_lock_timeout_ms));
   if (error){
      dev_err(pEnumFile->pEnumDevice->device, "%s: copy_from_user returned %ld\n", __FUNCTION__, error);
      goto bail;
   }

bail:
   return error;
}

long mda_ioctl_reglist(struct mda_device *mda_device, mda_io_buffer *user_io, int writeok)
{
   return mda_do_ioctl_reglist(&mda_device->device,
                                mda_device->parent->parent->parent,
                                &mda_device->device_address,
                                user_io,
                                writeok);
}

static long mda_ioctl_copyout_ui32(mda_io_buffer *user_io, ui32 dwData)
{
   long error;

   if (!user_io->output_buffer.user_address) return -EFAULT;
   if (user_io->output_buffer.user_size < sizeof(dwData)) return -EFAULT;
   error = copy_to_user(user_io->output_buffer.user_address, &dwData, sizeof(dwData));
   user_io->output_buffer.user_size = error ? 0 : sizeof(dwData);
   return error;
}

long mda_ioctl_version(mda_io_buffer *user_io, ui32 dwVersion)
{
   return mda_ioctl_copyout_ui32(user_io, dwVersion);
}

static ui32 aslenum_get_unique(struct aslenum_device *pEnumDevice)
{
   ui32 dwId = (ui32)pEnumDevice->pcidev->bus->number << 24;
   dwId += (ui32)pEnumDevice->pcidev->devfn << 16;
   return dwId;
}
 
long aslenum_ioctl_get_unique_id(mda_io_buffer *user_io, struct aslenum_device *pEnumDevice)
{
   return mda_ioctl_copyout_ui32(user_io, aslenum_get_unique(pEnumDevice));
}

long mda_ioctl_unique_id(mda_io_buffer *user_io, struct mda_device *pMdaDevice)
{
   ui32 dwId = aslenum_get_unique(pMdaDevice->parent->parent->parent);
   dwId += (ui32)pMdaDevice->channel << 8;
   dwId += (ui32)pMdaDevice->instance;
   return mda_ioctl_copyout_ui32(user_io, dwId);;
}

EXPORT_SYMBOL(mda_ioctl_reglist);
EXPORT_SYMBOL(mda_ioctl_version);
EXPORT_SYMBOL(mda_ioctl_unique_id);

