#include "precomp.h"

irqreturn_t aslenum_isr(int irq, void *data)
{
   irqreturn_t irqreturn = IRQ_NONE;
   struct aslenum_device *baseboard = (struct aslenum_device*)data;

   switch(baseboard->pcidev->device)
   {
   case FBD_V6_2PE8_DEVICE_ID:
   case FBD_A7_2PE4_DEVICE_ID:
   case FBD_KU_3PE4_DEVICE_ID:
   case FBD_KU_3PE8_DEVICE_ID: irqreturn = aslenum_firebird_isr(baseboard); break;

   case PHX_S3_1PEX_DEVICE_ID: irqreturn = aslenum_phoenix_isr(baseboard);  break;

   default: dev_err(baseboard->device, "!!!%s: Unrecognised baseboard in ISR!!!\n", __FUNCTION__); break;
   }

   return irqreturn;
}

static struct mda_irq_register * mda_get_irq_register_entry(struct mda_root_node *root_node, uint32_t irq_reg_offset)
{
   struct mda_irq_register *irq_reg_entry = NULL;

   list_for_each_entry(irq_reg_entry, &root_node->irq_register_list, this_mda_irq_register){
      if (irq_reg_entry->offset == irq_reg_offset) return irq_reg_entry;
   }
   return NULL;
}

static struct mda_irq * mda_get_irq_entry(struct mda_irq_register *irq_reg_entry, uint32_t irq_mask)
{
   struct mda_irq *irq_entry = NULL;

   list_for_each_entry(irq_entry, &irq_reg_entry->irq_function_list, this_mda_irq){
      if (irq_entry->mask == irq_mask) return irq_entry;
   }
   return NULL;
}

static int mda_add_irq_register_entry(struct mda_root_node *root_node, struct mda_irq_register *irq_reg)
{
   int error  = 0;

   INIT_LIST_HEAD(&irq_reg->irq_function_list);
   INIT_LIST_HEAD(&irq_reg->this_mda_irq_register);
   if (irq_reg->offset == MDA_INVALID_OFFSET) irq_reg->kvaddr = NULL;
   else irq_reg->kvaddr = root_node->parent->device_address.kvaddr + irq_reg->offset;
   irq_reg->value = 0;
   if (!mda_get_irq_register_entry(root_node, irq_reg->offset)){
      list_add_tail(&irq_reg->this_mda_irq_register, &root_node->irq_register_list);
   }
   else{
      dev_err(&root_node->device, "%s: Duplicate irq_reg_entry detected\n", __FUNCTION__);
      error = -EBUSY;
   }
   return error;
}

static int mda_add_irq_entry(struct mda_irq_register *irq_reg_entry, struct mda_irq *irq)
{
   int error = 0;

   if (!mda_get_irq_entry(irq_reg_entry, irq->mask)){
      list_add_tail(&irq->this_mda_irq, &irq_reg_entry->irq_function_list);
   }
   else{
      dev_err(&irq->mda_device->device, "%s: Duplicate irq_reg mask detected\n", __FUNCTION__);
      error = -EBUSY;
   }
   return error;
}

static int mda_del_irq_entry(struct mda_irq_register *irq_reg_entry, uint32_t mask)
{
   int error = 0;
   struct mda_irq *irq = NULL;
   struct mda_irq *tmp = NULL;

   list_for_each_entry_safe(irq, tmp, &irq_reg_entry->irq_function_list, this_mda_irq){
      if (irq->mask == mask){
         list_del(&irq->this_mda_irq);
         kfree(irq);
         goto bail;
      }
   }
   error = -ESRCH;
bail:
   return error;
}

static void mda_irq_reg_prune(struct mda_root_node *root_node)
{
   struct mda_irq_register *irq_reg = NULL;
   struct mda_irq_register *temp = NULL;

   list_for_each_entry_safe(irq_reg, temp, &root_node->irq_register_list, this_mda_irq_register){
      if (list_empty(&irq_reg->irq_function_list)){
         list_del(&irq_reg->this_mda_irq_register);
         kfree(irq_reg);
      }
   }
   return;
}

int mda_interrupt_register(struct mda_device *mda_device,
                           uint32_t irq_register,
                           uint32_t irq_mask,
                           irqreturn_t (*irq_handler)(struct mda_device *, void *),
                           void *data)
{
   int error = 0;
   struct mda_root_node * root_node = mda_device->parent->parent;
   struct mda_irq_register * existing_irq_reg_entry = NULL;
   struct mda_irq_register * irq_reg_entry = NULL;
   struct mda_irq * irq_entry = NULL;
   unsigned long saved_irq_flags = 0;

   spin_lock(&root_node->lock);

   existing_irq_reg_entry = mda_get_irq_register_entry(root_node, irq_register);
   if (existing_irq_reg_entry){
      irq_reg_entry = existing_irq_reg_entry;
   }
   else{
     irq_reg_entry = kzalloc(sizeof(*irq_reg_entry), GFP_KERNEL);
      if (irq_reg_entry == NULL){
          dev_err(&mda_device->device, "%s: kzalloc irq_reg failed\n", __FUNCTION__);
          error = -ENOMEM;
          goto bail;
      }
      irq_reg_entry->offset = irq_register;
      spin_lock_irqsave(&root_node->irq_lock, saved_irq_flags);
      error = mda_add_irq_register_entry(root_node, irq_reg_entry);
      spin_unlock_irqrestore(&root_node->irq_lock, saved_irq_flags);
      if (error){
         dev_err(&mda_device->device, "%s: failed to add irq_register_entry\n", __FUNCTION__);
         kfree(irq_reg_entry);
         irq_reg_entry = NULL;
         goto bail;
      }
   }

   irq_entry = kzalloc(sizeof(*irq_entry), GFP_KERNEL);
   if (!irq_entry){
      dev_err(&mda_device->device, "%s: kzalloc irq_entry failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto bail;
   }
   INIT_LIST_HEAD(&irq_entry->this_mda_irq);
   irq_entry->mask = irq_mask;
   irq_entry->mda_device = mda_device;
   irq_entry->mda_irq_handler = irq_handler;
   irq_entry->data = data;
   spin_lock_irqsave(&root_node->irq_lock, saved_irq_flags);
   error = mda_add_irq_entry(irq_reg_entry, irq_entry);
   spin_unlock_irqrestore(&root_node->irq_lock, saved_irq_flags);
   if (error){
      dev_err(&mda_device->device, "%s: mda_add_irq_entry failed\n", __FUNCTION__);
      goto bail;
   }

bail:
   if (error){
      spin_lock_irqsave(&root_node->irq_lock, saved_irq_flags);
      mda_irq_reg_prune(root_node);
      spin_unlock_irqrestore(&root_node->irq_lock, saved_irq_flags);
   }
   spin_unlock(&root_node->lock);

   return error;
}

int mda_interrupt_unregister(struct mda_device *mda_device,
                             uint32_t irq_register,
                             uint32_t irq_mask)
{
   int error = 0;
   struct mda_root_node * root_node = mda_device->parent->parent;
   struct mda_irq_register * irq_reg_entry = NULL;

   unsigned long saved_irq_flags = 0;

   spin_lock(&root_node->lock);

   irq_reg_entry = mda_get_irq_register_entry(root_node, irq_register);
   if (!irq_reg_entry){
      dev_err(&mda_device->device, "%s: no register entry found\n", __FUNCTION__);
      error = -ESRCH;
      goto bail;
   }
   spin_lock_irqsave(&root_node->irq_lock, saved_irq_flags);
   error = mda_del_irq_entry(irq_reg_entry, irq_mask);
   if (!error) mda_irq_reg_prune(root_node);
   spin_unlock_irqrestore(&root_node->irq_lock, saved_irq_flags);
   if (error){
      dev_err(&mda_device->device, "%s: no irq entry found\n", __FUNCTION__);
      goto bail;
   }

bail:
   spin_unlock(&root_node->lock);
   return (error);
}

EXPORT_SYMBOL(mda_interrupt_register);
EXPORT_SYMBOL(mda_interrupt_unregister);

